#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdio_ext.h>
#include <ctype.h>
#include <alloca.h>
#include <selinux/selinux.h>
#include <selinux/context.h>
#include "dso.h"

/* Define data structures */
typedef struct selabel {
	char* name;
	char* sename;
} selabel_t;

/* labellist is a simple linked list of selabels extracted from the translation file */
typedef struct labels {
	struct labels *next;
	selabel_t label;
} labels_t;

static labels_t *labellist=NULL;

void finish_context_translations(void) {
	labels_t *ptr=NULL;
	labels_t *current=NULL;
	if (!labellist) return;
	ptr=labellist;
	while(ptr) {
		if (ptr->label.name) free(ptr->label.name);
		if (ptr->label.sename) free(ptr->label.sename);
	        current=ptr;
		ptr=current->next;
		free(current);
	}
	labellist=NULL;
}

/* Process line from translation file. 
   Remove white space and set name do data before the "=" and sename to data
   after it
   May modify the data pointed to by the buffer paremeter */
static int process_label(char *buffer, labels_t **labels) {
	char *name, *name_end;
	labels_t *next=NULL;
	char *ptr, *tok;
	
	while(isspace(*buffer))
		buffer++;
	if(buffer[0] == '#') return 0;
	name = strtok_r(buffer, "=", &ptr);
	if (!name || !*name) return 0;
	tok = strtok_r(NULL, "\0", &ptr);
	if (!tok) return 0;
	name_end = tok - 2;
	while(isspace(*name_end))
	{
		name_end = '\0';
		name_end--;
	}
	while (isspace(*tok)) tok++;

	next=(labels_t *) calloc(1, sizeof(labels_t));
	if (!next) { 
		return -1;
	}

	next->label.name=strdup(name);
	if (!next->label.name) {
		free(next);
		return -1;
	}

	name_end = tok + strlen(tok) - 1;
	while(isspace(*name_end))
	{
		*name_end = '\0';
		name_end--;
	}
	next->label.sename=strdup(tok);
	if (!next->label.sename) {
		free(next->label.name);
		free(next);
		return -1;
	}
	*labels=next;
	return 1;
}

/* Look for selabel via internal name */
static const char *translate_internal(const char *labels) {
	labels_t *ptr=NULL;
	if (labellist) 
		for (ptr=labellist->next;ptr; ptr=ptr->next) 
			if (strcmp(labels,ptr->label.name)==0) {
				if ((ptr->label.sename) && 
				    (ptr->label.sename[0]==0))
					return NULL;
				else
					return ptr->label.sename; 
			}
	return labels;
}

static char *translate_categories_range(const char *sensitivity, 
					const char *first, 
					const char *last) {
	const char *trans=NULL;
	int begin=atoi(&first[1]);
	int end=atoi(&last[1]);
	char *ret=NULL;
	int i=0;
	for (i = begin; i<end+1; i++) {
		char *inbuf=NULL;
		if (asprintf(&inbuf,"%s:c%d", sensitivity, i) < 0) {
			return NULL;
		}	
		trans=translate_internal((const char *)inbuf);
		if (ret==NULL) {
			ret=strdup(trans);
		} else {
			char *tmp=ret;
			if (asprintf(&ret,"%s,%s", ret, trans) < 0) {
				free(inbuf);
				return NULL;
			}
			free(tmp);
		}
		free(inbuf); 
	}
	return ret;
}


static char *translate_categories(const char *sensitivity, const char *cat) {
	char *label=NULL;
	char *inbuf=NULL;
	int slen=strlen(sensitivity);
	char *ptr;
	int error=0;
	char *categories=strdupa(cat);
	char *tok=strtok_r(categories,",",&ptr);

	while (tok) {
		const char *trans;
		char *r;
		int mustfree=1;
		if ((r=strchr(tok,'.')) !=0) {
			trans=translate_categories_range(sensitivity, tok, ++r);
			if (trans == NULL) {
				error=1;
				break;
			}
			mustfree=1;
		}
		else {
			inbuf=malloc(slen+strlen(tok)+2);
			if (!inbuf) {
				error=1;
				break;
			}
			mustfree=0;
			sprintf(inbuf, "%s:%s", sensitivity, tok);
			trans=translate_internal(inbuf);
		}
		if (!label) {
			if ((label=strdup(trans)) == NULL) {
				error=1;
				break;
			}
		}
		else {
			char *tmp=label;
			if (asprintf(&label,"%s,%s", label, trans) < 0) {
				error=1;
				break;
			}
			free(tmp);
		}
		if (mustfree == 1) free((char *) trans); 
		free(inbuf); inbuf=NULL;
		tok=strtok_r(NULL,",",&ptr);
	}
	if (error) {
		free(label);
		free(inbuf);
		return NULL;
	}
	return label;
}


/* Look for selabel via internal name */
static const char *translate_label(const char *label) {
	char *translated=NULL;
	char *tmpbuf=NULL;
	char *ptr=NULL;
	char *tok=NULL;
	char *sensitivity=NULL;
	char *categories=NULL;
	/* Check for direct translation */
	const char *trans=translate_internal(label);
	if (! trans) return NULL; 
	if (strcmp(trans,label)!=0) return trans;

	tmpbuf=strdupa(label);
	tok=strtok_r(tmpbuf,":",&ptr);
	if (!tok) goto done;
	sensitivity=tok;
	tok=strtok_r(NULL,"\0",&ptr);
	if (!tok) goto done;
	categories=translate_categories(sensitivity, tok);
	if (categories) {
		const char *s=translate_internal(sensitivity);
		if (s)
			asprintf(&translated, "%s:%s", s, categories);
		else 
			translated=strdup(categories);
		
	}
 done:
	free(categories);
	return translated;
}
/* Look for selabel via internal name */
static const char *translate(const char *labels) {
	char *r;
	const char *trans=translate_internal(labels);
	if (! trans) return NULL; 
	if (strcmp(trans,labels)!=0) return strdup(trans);

	r=strchr(labels,'-');
	if (r) {
		int i=r-labels;
		const char *l1=NULL;
		const char *l2=NULL;
		char *ptr=strdup(labels);
		int error=0;
		ptr[i]=0;
		l1=translate_label(ptr);
		l2=translate_label(&ptr[i+1]);
		if (l1==NULL) l1=strdup(ptr);
		if (l2==NULL) l2=strdup(&ptr[i+1]);
		free(ptr);
		if (asprintf(&ptr,"%s-%s", l1, l2) < 0) error=1;
		free((char *) l1);
		free((char *) l2);
		if (error==0) return ptr;
		free(ptr);
		return NULL;
	} 
	else
		return translate_label(labels);
}

/* Look for selabel via external name */
static const char *untranslate_internal(const char *sename) {
	labels_t *ptr=NULL;
	if (labellist) 
		for(ptr=labellist->next;ptr; ptr=ptr->next) 
			if (strcmp(sename,ptr->label.sename)==0) 
				return ptr->label.name; 
	return sename;
}
static char *untranslate(const char *sename) {
	int error=0;
	const char *orig=untranslate_internal(sename);
	if (orig) {
		if (strcmp(orig,sename)==0) {
			char *untranslated=NULL;
			char *tmpbuf=strdupa(orig);
			char *ptr=NULL;
			char *tok=NULL;
			char *sensitivity=NULL;
			const char *category=NULL;
			tok=strtok_r(tmpbuf,",",&ptr);
			while (tok) {
				const char *utrans;
				int idx;
				utrans=untranslate_internal(tok);
				if (strcmp(utrans, tok)==0) {
					error=1;
					break;
				}
				idx = strchr(utrans,':') - utrans;
				if (sensitivity) {
					if (strncmp(sensitivity, utrans, strlen(sensitivity)) != 0) {
						error=1;
						break;
					}
				} else {
					sensitivity = alloca(idx + 1);
					sensitivity[idx] = '\0';
					strncpy(sensitivity, utrans, idx);
				}
				category=&utrans[idx+1];
				if (!untranslated) {
					if (asprintf(&untranslated,"%s:%s", sensitivity, category) < 0) {
						error=1;
						break;
					}
				}
				else {
					char *tmp=untranslated;
					if (asprintf(&untranslated,"%s,%s", untranslated, category) < 0) {
						error=1;
						break;
					}
					free(tmp);
				}
				tok=strtok_r(NULL,",",&ptr);
			}
		done:
			if (error) {
				free(untranslated);
				return strdup(orig);
			}
			else
				return untranslated;
		} 
		else 
			return strdup(orig);
	}
	return strdup(sename);
}

/* Read in translation file Only runs once per process.  
   Might want to change to some kind of reload eventually, for long running
   processes.
 */
int init_context_translations(void) {
	FILE *cfg=NULL;
	labels_t *ptr=NULL;
	labels_t *next=NULL;
	size_t size=0;
	char *buffer=NULL;

	if (is_selinux_mls_enabled() <= 0 ) {
		return 1;
	}

	cfg = fopen(selinux_translations_path(),"r");
	if (!cfg) return 1;

	__fsetlocking(cfg, FSETLOCKING_BYCALLER);
	ptr=labellist=calloc(1,sizeof(labels_t));
	if (!ptr) {
		fclose(cfg);
		return 1;
	}

	while (getline(&buffer, &size, cfg) > 0) {
		if(process_label(buffer, &next)) {
			ptr->next=next;
			ptr=next;
			if ((strcasecmp(next->label.name,"disable")==0) &&
			    (strcmp(next->label.sename,"1") == 0)) {
				finish_context_translations();
				break;
			}
		}
	}
	free(buffer);

	fclose(cfg);
	return 0;
}

/* Look for selabel via external name */
int translate_context( const security_context_t oldcon, security_context_t *rcon) {
	const char *range=NULL;
	context_t con=context_new(oldcon);

	if (! con) return -1;

	if (labellist) {
		range=context_range_get(con);
		if (range) {
			const char *ptr=translate(range);
			context_range_set(con,ptr);
			free((char *)ptr);
		}
	}
	*rcon=strdup(context_str(con));
	context_free(con);
	return 0;
}

/* Look for selabel via external name */
int untranslate_context( const security_context_t oldcon, security_context_t *rcon) {
	const char *range=NULL;
	char *newrange=NULL;
	context_t con=context_new(oldcon);

	if (! con) return -1;
	if (labellist) {
		range=context_range_get(con);
		if (range) {
			char *ptr=strchr(range,'-');
			/* if ranged label */
			if(ptr) {
                                /* Now tmpbuf has minimum level string
				 * and &range[i] has clearance level string.
				 * If minimum and clearance level is same 
				 * use single level string as range.
				 * This removes the need to define same 
				 * level ranges (ex: SystemHigh-SystemHigh)
				 * in translation table */
				size_t i=ptr-range;
				ptr++;
				if ((i==strlen(ptr)) && 
				    (strncmp(range, ptr, i)==0)) {
						newrange=untranslate(ptr);
				} else {
					newrange=untranslate(range);
				}
			} 
			else 
				newrange=untranslate(range);
		}
		else 
			newrange=untranslate("");
		if (newrange) {
			context_range_set(con,newrange);
			free(newrange);
		}
	}
	*rcon=strdup(context_str(con));
	context_free(con);
	return 0;
}


